;;;-*- Mode: Lisp; Package: ccl -*-
;*********************************************************************
;*                                                                   *
;*        T E X T U R E   and   I M A G E   I M P O R T              *
;*                                                                   *
;*********************************************************************
   ;* Author    : Alexander Repenning (alexander@agentsheets.com)    *
   ;*             http://www.agentsheets.com                         *
   ;* Copyright : (c) 1996-2002, AgentSheets Inc.                    *
   ;* Filename  : Texture-and-Image-import.lisp                      *
   ;* Updated   : 11/29/02                                           *
   ;* Version   :                                                    *
   ;*   1.0     : 11/29/02 from MacOpenGL 0.8.2                      *
   ;* SW/HW     : PowerPC G4, MCL 5                                  *
   ;* Abstract  : QuickTime based Texture and Image import functions *
   ;*                                                                *
   ;******************************************************************

(in-package :ccl)

(use-package :gl)


(defun ARGB-TO-RGBA-BUFFER (Buffer) "
  in/out: Buffer.
  Rearange buffer from ARGB to RGBA order."
  ;; not the fast way but usually this is only done just once as part of reading the texture from file
  ;; could speed up with #_blockMove
  (dotimes (i (truncate (#_GetPtrSize Buffer) 4))
    (let* ((Addr (* i 4))
           (Alpha (%get-byte Buffer Addr)))
      (%put-byte Buffer (%get-byte Buffer (+ Addr 1)) (+ Addr 0))
      (%put-byte Buffer (%get-byte Buffer (+ Addr 2)) (+ Addr 1))
      (%put-byte Buffer (%get-byte Buffer (+ Addr 3)) (+ Addr 2))
      (%put-byte Buffer Alpha (+ Addr 3)))))


(defun FLIP-VERTICAL-BUFFER (Buffer Rowsize) "
  in/out: Buffer {bytes}.
  in: Rowsize {int}. number of bytes
  Flip content of buffer vertically"
  ;; a shame that this is necessary but this is how OpenGL wants its buffer
  ;; should be fairly fast: rows * 3 blockmoves
  (unless (zerop (mod (#_GetPtrSize Buffer) Rowsize))
    (error "cannot flip buffer: rowsize inconsistent with buffersize"))
  (let ((Row-Buffer (#_NewPtr Rowsize))
        (Rows (truncate (#_GetPtrSize Buffer) Rowsize)))
    ;; outside-in swap top with bottom until reaching middle
    (dotimes (Row (truncate Rows 2))
      ;; 1) row from above into row buffer
      (#_BlockMove
       (%inc-ptr Buffer (* Row Rowsize))
       Row-Buffer
       Rowsize)
      ;; 2) row from below into row above 
      (#_BlockMove
       (%inc-ptr Buffer (* (- Rows Row 1) Rowsize))
       (%inc-ptr Buffer (* Row Rowsize))
       Rowsize)
      ;; 2) row buffer into row below 
      (#_BlockMove
       Row-Buffer
       (%inc-ptr Buffer (* (- Rows Row 1) Rowsize))
       Rowsize))
    (#_DisposePtr Row-Buffer)))
      

(defun CREATE-TEXTURE-FROM-FILE (Filename &key Verbose) "
  in:  Filename {string}, &key Verbose {boolean}.
  out: OpenGL-Texture-Name {int}.
  Load the <Filename> texture inside the texture directory.
  - File must be 32 bit ARGB compatible, e.g., .png with mask or 24 bit RGB
  - Image size must be 2^n x 2^m, at least 64 x 64
  - This function must be called with active AGL context, e.g., inside OpenGL-Window INIT method."
  (rlet ((&gworldPtr :pointer)
         (&gi :handle)
         (Fsspec :fsspec))
    ;; Set File Spec
    (with-pstrs ((Name (mac-namestring (probe-file Filename))))
      (unless (zerop (#_fsmakefsspec 0 0 Name Fsspec))
        (error "Invalid file reference ~A" Filename))
      ;; Read Image dimensions and create GWorld
      (unless (zerop (#_GetGraphicsImporterForFile Fsspec &gi))
        (error "_GetGraphicsImporterForFile")))
    (let ((gi (%get-ptr &gi)))          ;; with-dereferenced-handles does not work here in OS X
      (rlet ((&ImageDescriptor :pointer))
        (#_GraphicsImportGetImageDescription gi &ImageDescriptor)
        (let ((ImageDescriptor (%get-ptr &ImageDescriptor)))
          (let ((Width (rref ImageDescriptor :imagedescription.width))
                (Height (rref ImageDescriptor :imagedescription.height))
                (Depth (rref ImageDescriptor :imagedescription.depth))
                (Bytes-per-pixel (truncate (rref ImageDescriptor :imagedescription.depth) 8)))
            (when Verbose (format t "~%-loading texture ~A size: ~A x ~A, depth: ~A" Filename Width Height Depth))
            ;; create the GWorld
            (let ((&texture (#_newPtr (* Width Height Bytes-per-pixel))))
              (rlet ((&rect :rect :topleft #@(0 0) :botright (make-point Width Height)))
                (unless (zerop (#_QTNewGWorldFromPtr &gworldPtr (ecase Bytes-per-pixel 
                                                                  (4 #$k32ARGBPixelFormat)
                                                                  (3 #$k24RGBPixelFormat))
                                &rect (%null-ptr) (%null-ptr) 0 &texture (* Width Bytes-per-pixel)))
                  (error "Cannot create GWorld from texture file"))
                (let* ((gworldPtr (%get-ptr &gworldPtr))
                       (Pixmap-Handle (#_getGWorldPixmap gworldPtr)))
                  (#_LockPixels Pixmap-Handle)
                  ;; set port and draw image into texture buffer
                  (#_GraphicsImportSetGWorld gi gworldPtr (%null-ptr))
                  (#_GraphicsImportDraw gi)
                  (setq &texture (#_GetPixBaseAddr Pixmap-Handle))
                  ;; make buffer organization compatible with OpenGL
                  (when (= Bytes-per-pixel 4) (argb-to-rgba-buffer &texture))
                  (flip-vertical-buffer &texture (* Bytes-per-pixel Width))
                  ;; create the OpenGL texture (hopefully in accelerator memory) and define parameters
                  (rlet ((&texName :long))
                    (glGenTextures 1 &texName)
                    (glBindTexture gl_texture_2d (%get-long &texName))
                    (gltexparameteri gl_texture_2d gl_texture_wrap_s gl_repeat)
                    (gltexparameteri gl_texture_2d gl_texture_wrap_t gl_repeat)
                    ;; make textures look smooth (can be costly)
                    (gltexparameteri gl_texture_2d gl_texture_mag_filter gl_linear)
                    (gltexparameteri gl_texture_2d gl_texture_min_filter gl_linear)
                    (let ((Format (ecase Bytes-per-pixel (4 gl_rgba) (3 gl_rgb))))
                      (glTexImage2D gl_texture_2d 0 Format Width Height 0 Format gl_unsigned_byte &texture))
                    ;; texture is in OpenGL now, can get rid of all the support data structures
                  (#_UnlockPixels Pixmap-Handle)
                  (#_CloseComponent gi)
                  (#_DisposeGWorld gworldPtr)
                  (#_DisposePtr &texture)
                  (%get-long &texName)))))))))))


(defun CREATE-IMAGE-FROM-FILE (Filename) "
  in:  Filename {string}.
  out: Pixels {imageBuffer}.
  Create an image buffer from <Filename>
  - File must be 32 bit ARGB compatible, e.g., .png with mask or 24 bit RGB."
  (rlet ((&gworldPtr :pointer)
         (&gi :handle)
         (Fsspec :fsspec))
    ;; Set File Spec
    (with-pstrs ((Name (mac-namestring (probe-file (make-pathname :directory *Image-Directory* 
                                                                  :name Filename)))))
      (unless (zerop (#_fsmakefsspec 0 0 Name Fsspec))
        (error "Invalid file reference ~A" Filename))
      ;; Read Image dimensions and create GWorld
      (unless (zerop (#_GetGraphicsImporterForFile Fsspec &gi))
        (error "_GetGraphicsImporterForFile")))
    (let ((gi (%get-ptr &gi)))          ;; with-dereferenced-handles does not work here in OS X
      (rlet ((&ImageDescriptor :pointer))
        (#_GraphicsImportGetImageDescription gi &ImageDescriptor)
        (let ((ImageDescriptor (%get-ptr &ImageDescriptor)))
          (let ((Width (rref ImageDescriptor :imagedescription.width))
                (Height (rref ImageDescriptor :imagedescription.height))
                (Depth (rref ImageDescriptor :imagedescription.depth))
                (Bytes-per-pixel (truncate (rref ImageDescriptor :imagedescription.depth) 8)))
            (format t "~%-loading image ~A size: ~A x ~A, depth: ~A" Filename Width Height Depth)
            ;; create the GWorld
            (let ((&texture (#_newPtr (* Width Height Bytes-per-pixel))))
              (rlet ((&rect :rect :topleft #@(0 0) :botright (make-point Width Height)))
                (unless (zerop (#_QTNewGWorldFromPtr &gworldPtr (ecase Bytes-per-pixel 
                                                                  (4 #$k32ARGBPixelFormat)
                                                                  (3 #$k24RGBPixelFormat))
                                &rect (%null-ptr) (%null-ptr) 0 &texture (* Width Bytes-per-pixel)))
                  (error "Cannot create GWorld from texture file"))
                (let* ((gworldPtr (%get-ptr &gworldPtr))
                       (Pixmap-Handle (#_getGWorldPixmap gworldPtr)))
                  (#_LockPixels Pixmap-Handle)
                  ;; set port and draw image into texture buffer
                  (#_GraphicsImportSetGWorld gi gworldPtr (%null-ptr))
                  (#_GraphicsImportDraw gi)
                  (setq &texture (#_GetPixBaseAddr Pixmap-Handle))
                  ;; make buffer organization compatible with OpenGL
                  (when (= Bytes-per-pixel 4) (argb-to-rgba-buffer &texture))
                  (flip-vertical-buffer &texture (* Bytes-per-pixel Width))
                  (#_UnlockPixels Pixmap-Handle)
                  (#_CloseComponent gi)
                  (#_DisposeGWorld gworldPtr)
                  &texture)))))))))

