;;===========================================================================
;OM API 
;Multiplatform API for OpenMusic
;Macintosh version (Digitool Macintosh Common Lisp - MCL)
;
;Copyright (C) 2004 IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Augusto Agon and Jean Bresson
;;===========================================================================

;;===========================================================================
;DocFile
; GRAPHIC STRUCTURES (points, rects, regions, colors, fonts) 
; AND DRAWING FUNCTIONS 
;;===========================================================================



(in-package :om-api)

;;;=========================
;;; export :
;;;=========================
(export '(
                om-make-point
                om-make-big-point
                om-point-p
                om-points-equal-p
                om-add-points
                om-subtract-points
                om-add-big-points
                om-subtract-big-points
                om-point-h
                om-point-v
                
                om-make-rect
                om-pts-to-rect
                om-sect-rect
                om-rect-empty
                om-rect-top
                om-rect-left
                om-rect-bottom
                om-rect-right
                om-rect-w
                om-rect-h
                om-rect-topleft
                om-rect-bottomright
                om-point-in-rect-p
                
                om-open-region
                om-close-region
                om-open-region-add-rect
                om-open-region-add-line
                om-new-region
                om-dispose-region
                om-region-add-rect
                om-region-add-line
                om-set-rect-region
                om-union-region
                om-point-in-region-p
                om-copy-region
                om-difference-region
                om-paint-region
                om-offset-region
                
                om-get-current-port
                
                om-with-focused-view
                om-with-clip-rect
                om-with-bg-color
                om-with-fg-color
                
                om-draw-char
                om-draw-string
                om-draw-line
                om-draw-point
                om-draw-rect
                om-draw-round-rect
                om-fill-round-rect
                om-fill-rect
                om-draw-ellipse
                om-fill-ellipse
                om-draw-ellipse-arc
                
                om-erase-rect-content
                om-erase-rect
                om-erase-line
                
                *om-select-color*
                om-draw-selection-rect
                om-draw-hilite-rect

                om-draw-rect-outline
                
                *om-black-pattern*
                *om-darkgray-pattern*
                *om-gray-pattern*
                *om-lightgray-pattern*
                *om-white-pattern*
                
                om-with-pen
                om-with-line-size
                om-with-dashline
                om-with-hilite-foreground
                
                om-make-color
                om-color-p
                om-color-r
                om-color-g
                om-color-b
                *om-black-color* 
                *om-dark-gray-color* 
                *om-gray-color* 
                *om-light-gray-color* 
                *om-white-color* 
                *om-window-def-color*
                
                *om-default-font1*
                *om-default-font2*
                *om-default-font3*
                *om-default-font4*
                
                *om-default-font1b*
                *om-default-font2b*
                *om-default-font3b*
                *om-default-font4b*
	        *om-score-font*
                *om-verdana-font*
                *om-def-font-face*

                om-make-font
                om-font-p
                om-font-face
                om-font-family
                om-font-size
                om-font-mode
                om-font-style
                om-string-size
                om-string-h
                om-with-font
                
                om-correct-point
                om-correct-font
                om-correct-color

                ) :om-api)




;;;=========================
;;;COMPAT TYPES
;;;=========================
(defun om-correct-color (color) color)
(defun om-correct-point (point) point)
(defun om-correct-font (font)  font)


;;;=========================
;;;POINTS
;;;=========================
(defclass ompoint () 
  ((px :accessor px :initarg :px :initform 0)
   (py :accessor py :initarg :py :initform 0)))

(defmethod make-load-form ((self ompoint) &optional env)
  (declare (ignore env))
  `(make-instance 'ompoint :px ,(px self) :py ,(py self)))

(defun om-make-point (x y) 
  (make-instance 'ompoint :px x :py y)) 

(defmethod om-point-p ((self t)) nil)
(defmethod om-point-p ((self ompoint)) t)

(defun om-make-big-point (x y)
  (make-instance 'ompoint :px x :py y))


(defmethod om-point-h ((point ompoint))
  (px point))


(defmethod om-point-v ((point ompoint))
  (py point))


(defun om-add-points (point1 point2)
   (om-make-point (+ (om-point-h point1) (om-point-h point2))
                  (+ (om-point-v point1) (om-point-v point2))))


(defun om-subtract-points (point1 point2)
   (om-make-point (- (om-point-h point1) (om-point-h point2))
                  (- (om-point-v point1) (om-point-v point2))))



(defun om-add-big-points (point1 point2)
  (om-add-points point1 point2))


(defun om-subtract-big-points (point1 point2)
  (om-subtract-points point1 point2) )


(defun om-points-equal-p (point1 point2)
  (and (equal (px point1) (px point2)) 
       (equal (py point1) (py point2))))

(defmethod print-point ((point ompoint))
  (let ((str (format nil "point>>> h:~A v:~A" (om-point-h point) (om-point-v point))))
    (print str))
  point)


;;;=========================
;;; RECTS 
;;;=========================


(defclass om-rect ()
  ((rx :accessor rx :initarg :rx :initform 0)
   (ry :accessor ry :initarg :ry :initform 0)
   (rw :accessor rw :initarg :rw :initform 0)
   (rh :accessor rh :initarg :rh :initform 0)))

(defun om-make-rect (left top right bottom)
(make-instance 'om-rect
	       :rx left
	       :ry top
	       :rw (- right left)
	       :rh (- bottom top)))

(defun om-pts-to-rect (ompoint1 ompoint2)
 (let ((left (min (om-point-h ompoint1)(om-point-h ompoint2)))
        (top (min (om-point-v ompoint1)(om-point-v ompoint2)))  
        (right (max (om-point-h ompoint1)(om-point-h ompoint2)))
        (bottom (max (om-point-v ompoint1)(om-point-v ompoint2))))
   (make-instance 'om-rect
		:rx left
		:ry top
		:rw (- right left)
		:rh (- bottom top))))
 
(defun om-rect-top (rect)
  (ry rect))

(defun om-rect-bottom (rect)
  (+ (ry rect) (rh rect)))  

(defun om-rect-left (rect)
  (rx rect))

(defun om-rect-right (rect)
  (+ (rx rect) (rw rect)))

(defun om-rect-w (rect)
  (rw rect))

(defun om-rect-h (rect)
  (rh rect))

(defun om-rect-topleft (rect)
  (om-make-point (rx rect) (ry rect)))

(defun om-rect-bottomright (rect)
  (om-make-point (+ (rx rect) (rw rect)) (+ (ry rect) (rh rect))))


(defmethod rect-p ((self t)) nil)
(defmethod rect-p ((self om-rect)) t)

(defun om-sect-rect (rect1 rect2)
  (let* ((tx (max (rx rect1) (rx rect2)))
	 (ty (max (ry rect1) (ry rect2)))
	 (t2x (min (+ (rx rect1) (rw rect1))  (+ (rx rect2) (rw rect2))))
	 (t2y (min (+ (ry rect1) (rh rect1))  (+ (ry rect2) (rh rect2)))))
    (if (or (< t2x tx) (< t2y ty))
	nil
	(om-make-rect tx ty t2x t2y))))

(defun om-rect-empty (rect)
  (equal rect nil))

(defun om-point-in-rect-p (point rect)
  (and (and (>= (px point) (rx rect))
	    (<= (px point) (+ (rx  rect) (rw rect))))
       (and (>= (py point) (ry rect))
	    (<= (py point) (+ (ry rect) (rh rect))))))

(defun om-copy-rect (rect)
  (make-instance 'om-rect
	       :rx (rx rect)
	       :ry (ry rect)
	       :rw (rw rect)
	       :rh (rh rect)))



;;;=========================
;;;REGIONS 
;;;=========================
(defclass om-region () 
  ((viewport :initform nil :initarg :viewport :accessor viewport)
   (elements :initform nil :initarg :elements :accessor elements)))

;;;====================================
;;; 2 MANIERES DE DEFINIR DES REGIONS... peut tre  unifier...

;;; ====================================
;;; ouvrir la region, ajouter des elements, et fermer (close retourne la region) :

(defvar *current-region* nil)

(defun om-open-region (view)
  (setf *current-region* (make-instance 'om-region :viewport view))) 

(defun om-close-region (view)
   (declare (ignore view))
   *current-region*)

(defun om-open-region-add-rect (left top right bottom)
   (if *current-region* (push (om-make-rect left top right bottom) (elements *current-region*))))

(defun om-open-region-add-line (p1 p2 &optional (approx nil))
   (if *current-region* (push (list p1 p2 approx) (elements *current-region*))))

(defun om-region-bounds ( region)
  (let ((minx 10000000000000000) ;ayay
        (miny 10000000000000000)
        (maxx -1000000000000000)
        (maxy -1000000000000000))
   (loop for elt in (elements region) do
             (if (rect-p elt) 
                 (progn
                   (setf minx (min minx (rx elt)))
                   (setf miny (min miny (ry elt)))
                   (setf maxx (max maxx (+ (rx elt) (rw elt))))
                   (setf maxy (max maxy (+ (ry elt) (rh elt)))))
               (progn
                   (setf minx (min minx (om-point-h (first elt))))
                   (setf miny (min miny (om-point-v (first elt))))
                   (setf maxx (max maxx (om-point-h (second elt))))
                   (setf maxy (max maxy (om-point-v (second elt)))))))
   (list minx miny maxx maxy)))


;;; ====================================
;;; methode OO

(defun om-new-region ()
   (make-instance 'om-region))

(defmethod om-dispose-region ((self om-region))
   (setf self nil))

;;; not used
(defmethod om-region-add-rect ((self om-region) left top right bottom)
   (push (om-make-rect left top right bottom) (elements self)))

;;; not used
(defmethod om-region-add-line ((self om-region) p1 p2 &optional (approx nil))
   (push (list p1 p2 approx) (elements self)))

(defun om-set-rect-region (region left top right bottom)
     (setf (elements region) (list (om-make-rect left top right bottom)))
     region)


;;; =====================================
;;; Test region

(defun om-union-region (reg1 reg2)
   (let ((reg-union (make-instance 'om-region)))
     (setf (elements reg-union) (append (elements reg1) (elements reg2)))
     reg-union))

(defmethod om-point-in-region-p ((self om-region) where)
   (let ((in nil))
     (loop for partie in (elements self)
            while (not in) do
               (cond ((rect-p partie) (if (om-point-in-rect-p where partie) (setf in t)))
                           (t (let ((p1 (first partie)) (p2 (second partie)) (approx (third partie)))
                                (unless (om-points-equal-p p1 p2)  
                                   (if approx
                                       (if (gp::point-in-line-hull (om-point-h p1) (om-point-v p1)
                                                                   (om-point-h p2) (om-point-v p2)
                                                                   (om-point-h where) (om-point-v where) approx)
                                           (setf in t))
                                     (if (gp::point-in-line-hull (om-point-h p1) (om-point-v p1)
                                                                   (om-point-h p2) (om-point-v p2)
                                                                   (om-point-h where) (om-point-v where) ) (setf in t)))
                                     )))))
     in))


(defun om-copy-region (region) 
   (let ((r (om-new-region)))
     (setf (elements r) 
        (loop for elt in (elements region) collect
                  (if (rect-p elt) (om-copy-rect elt)
                    (copy-list elt))))
     r))



;;; A FAIRE 
(defun om-difference-region (r1 r2 r3) r1)


(defun om-paint-region (view region)
   (loop for elt in (elements region) do
             (if (rect-p elt) 
                 (gp:draw-rectangle (om-get-view view) (rx elt) (ry elt) (rw elt) (rh elt) :filled nil)
               (gp::draw-line (om-get-view view) (om-point-h (first elt)) (om-point-v (first elt)) (om-point-h (second elt)) (om-point-v (second elt))))
             ))
             

(defun om-offset-region (region h v) 
   (let ((newelts (loop for elt in (elements region) collect 
                                    (if (rect-p elt) (om-make-rect (+ (rx elt) h) (+ (ry elt) v)
                                                                   (+ (rx elt) (rw elt)  h)
                                                                   (+ (ry elt) (rh elt) v))
                                      (list (om-add-point (first elt) (om-make-point h v))
                                             (om-add-points (second elt) (om-make-point h v))
                                             (third elt))))))
     (setf (elements region) newelts)
     region))





;;;=========================
;;;COLORS
;;;=========================



(defmacro om-make-color (r g b)
  `(color:make-rgb ,r ,g ,b))

(defmethod om-color-p ((self t)) (color::color-spec-p self))

(defun om-color-r (color)
  (color::color-red color))

(defun om-color-g (color)
  (color::color-green color))

(defun om-color-b (color)
   (color::color-blue color))


;;; system colors :
(defvar *om-black-color* (color::get-color-spec :black))
(defvar  *om-light-gray-color* (om-make-color 0.75 0.75 0.75) )
(defvar *om-gray-color* (om-make-color 0.5 0.5 0.5))
(defvar *om-dark-gray-color* (om-make-color 0.25 0.25 0.25))
(defvar *om-white-color* (color::get-color-spec :white))

(defvar *om-window-def-color* nil)
; (setf *om-window-def-color* (om-make-color 0.803 0.854 0.855))
(setf *om-window-def-color* :transparent)

;(om-choose-color-dialog :color (om-make-color 0.803 0.854 0.855))


;;;=========================
;;;FONTS
;;;=========================

;;; view bidon
(defvar *record-view* nil)

(defun init-record-view ()
  (let* ((pl (make-instance 'capi:pinboard-layout))
         (win (capi:display (make-instance 'capi:interface 
                                           :display-state :hidden
                                           :layout pl))))
    (setf  *record-view* pl)))

(om-add-init-func 'init-record-view)

(defmethod om-font-p ((self t)) (gp::font-description-p self))

;(defvar *om-font-modes* '(:srcCopy :srcOr :srcXor :srcBic :srcPatCopy :srcPatOr :srcPatXor :srcPatBic))
;(defvar *om-font-styles* '(:plain :bold :condense :extend :italic :outline :shadow :underline))

(defun om-make-font (face size &key (family nil) (style nil) (mode nil))
  (declare (ignore mode))
  (gp::make-font-description 
   :name face
   :size (round size)
   :family (or family face)
   :slant (if (member :italic style) :italic :roman)
   :weight (if (member :bold style) :bold :regular)
   ))

;(gp::font-description-attributes (gp::make-font-description :family "Times" :size 14))
;(setf fo (capi::prompt-for-font nil))

(defun om-font-face (font) 
  (gp::font-description-attribute-value font :name))

(defun om-font-size (font) 
   (gp::font-description-attribute-value font :size))

(defun om-font-family (font) 
  (gp::font-description-attribute-value font :family))

(defun om-font-style (font) 
  (cond ((and (equal (gp::font-description-attribute-value font :weight) :bold)
              (equal (gp::font-description-attribute-value font :slant) :italic))
         '(:bold :italic))
        ((equal (gp::font-description-attribute-value font :slant) :italic) 
         '(:italic))
        ((equal (gp::font-description-attribute-value font :weight) :bold)
         '(:bold))
        (t '(:plain))))

(defun om-font-mode (font) nil)

(defun om-string-size (name &optional (font 6))
  (setf (capi::simple-pane-font *record-view*) font)
  (if name (gp::port-string-width *record-view* name) 0)
  ;(if name (GP:GET-STRING-EXTENT *record-view* name) 0)
)

(defun om-string-h (&optional (font *om-default-font2*))
  (setf (capi::simple-pane-font *record-view*) font)
  (multiple-value-bind (x1 y1 x2 y2) (gp::get-string-extent *record-view* "ABC") (- y2 y1)))

; (om-string-size "Hello" (om-make-font "arial" 20))
; GP:GET-STRING-EXTENT


(defvar *om-score-font* nil)
(setf *om-score-font* "Times New Roman" )

(defvar *om-verdana-font* nil)
(setf *om-verdana-font* "Verdana" )

(defvar *om-def-font-face* nil)
(setf *om-def-font-face* "Verdana")

(defvar *om-default-font1* nil)
(defvar *om-default-font2* nil)
(defvar *om-default-font3* nil)
(defvar *om-default-font4* nil)

(defvar *om-default-font1b* nil)
(defvar *om-default-font2b* nil)
(defvar *om-default-font3b* nil)
(defvar *om-default-font4b* nil)

(setf *om-default-font1* (om-make-font *om-def-font-face* 10))
(setf *om-default-font2* (om-make-font *om-def-font-face* 12))
(setf *om-default-font3* (om-make-font *om-def-font-face* 14))
(setf *om-default-font4* (om-make-font *om-def-font-face* 18))

(setf *om-default-font1b* (om-make-font "Verdana-Bold" 10 :family "Verdana" :style '(:bold)))
(setf *om-default-font2b* (om-make-font "Verdana-Bold" 12 :family "Verdana" :style '(:bold)))
(setf *om-default-font3b* (om-make-font "Verdana-Bold" 14 :family "Verdana" :style '(:bold)))
(setf *om-default-font4b* (om-make-font "Verdana-Bold" 18 :family "Verdana" :style '(:bold)))




;====== DRAW =====
;==========================================

(defvar *curstream* nil)

(defmacro om-with-focused-view (view &body body)
  `(let ((*curstream* (om-get-view ,view)))
     (when *curstream*
     ,@body)))


(defun om-get-current-port ()
  *curstream*)


(defmacro om-with-font (font &rest body)
  `(gp::with-graphics-state (*curstream* :font (if (gp::font-description-p ,font)
                                                   (gp::find-best-font *curstream* ,font)
                                                 ,font))
     ,@body))

;non
(defmacro om-with-clip-rect (view rect &body body)
  `(let () ,@body))

;non
(defun om-set-clip-rect (view rect)
  (clip-rect view (om-rect-left rect) (om-rect-top rect)  (om-rect-bottom rect) (om-rect-right rect)))

(defmacro om-with-fg-color (view color &body body)
  `(let ((port (if ,view (om-get-view ,view) *curstream*)))
     (gp::with-graphics-state (port :foreground ,color)
       ,@body)))

(defmacro om-with-bg-color (view color &body body)
  `(let ((port (if ,view (om-get-view ,view) *curstream*)))
     (gp::with-graphics-state (port :background ,color)
       ,@body)))

;-----------------------------------
(defun om-draw-char (x y cn)
  (gp:draw-character *curstream* cn x y ))

(defun om-draw-string (x y str)
  (gp:draw-string *curstream* str x y ))

(defun om-draw-line (x1 y1 x2 y2)  
  (gp:draw-line *curstream* x1 y1 x2 y2 ))


(defun om-draw-point (x y)  
   (gp:draw-point *curstream* x y))

(defun om-draw-rect (x &optional y (w  0) (h 0))
  (let (top left wi he)
    (if y
        (setf left x top y wi w he h)
      (setf left (rx x) top (ry x) wi (rw x) he (rh x)))
    (gp:draw-rectangle *curstream* left top wi he :filled nil)))

(defun om-fill-rect (x &optional y (w 0) (h 0)) 
  (let (top left wi he)
  (if y
      (setf left x top y wi w he h)
    (setf left (rx x) top (ry x) wi (rw x) he (rh x)))
  (gp:draw-rectangle *curstream* left top wi he :filled t)))


;il faut les enlever
(defun om-draw-round-rect (x y w h ow oh) 
  t)

(defun om-fill-round-rect (x y w h ow oh) 
  t)

(defun om-draw-ellipse (x y w h)  
  (gp:draw-ellipse *curstream* x y w h :filled nil))

(defun om-fill-ellipse (x y w h)  
   (gp:draw-ellipse *curstream* x y w h :filled t))

;a faire
(defun om-draw-ellipse-arc (center maj-axe min-axe major-angle start-angle length-angle) t)


;------------
(defun om-erase-rect-content (x &optional y (w 0) (h 0))
  (let (top left wi he)
    (if y (setf left x top y wi w he h)
      (setf left (rx x) top (ry x) wi (rw x) he (rh x)))
    (gp:draw-rectangle *curstream* left top wi he :filled t 
                       :foreground (simple-pane-background *curstream*))))

(defun om-erase-rect (x &optional y (w 0) (h 0))
  (let (top left wi he)
    (if y (setf left x top y wi w he h)
      (setf left (rx x) top (ry x) wi (rw x) he (rh x)))
    (gp:draw-rectangle *curstream* left top wi he :filled nil 
                       :foreground (simple-pane-background *curstream*))))


(defun om-erase-line (x1 y1 x2 y2) 
  (gp:draw-line *curstream* x1 y1 x2 y2
                :foreground (simple-pane-background *curstream*)))

;------------


;;;=========================
;;; PATTERNS DE SURFACE
;;;=========================

(defvar *om-black-pattern* nil)
(defvar *om-darkgray-pattern* nil)
(defvar *om-gray-pattern* nil)
(defvar *om-lightgray-pattern* nil)
(defvar *om-white-pattern* nil)

(defun om-init-patterns ()
  (setf *om-black-pattern* *black-pattern*)
  (setf *om-darkgray-pattern* *dark-gray-pattern*)
  (setf *om-gray-pattern* *gray-pattern*)
  (setf *om-lightgray-pattern* *light-gray-pattern*)
  (setf *om-white-pattern* *white-pattern*))

;(om-add-init-func 'om-init-patterns)


(defvar *om-select-color* nil)
(setf *om-select-color* (color::make-rgb 0.3 0.45 0.75 1))

(defvar *om-select-color-alpha* nil)
(setf *om-select-color-alpha* (color::make-rgb 0.2 0.35 0.6 0.2))

(defun om-draw-selection-rect (x y w h &key (mode :xor))
  (let ((left (if (minusp w) (+ x w) x))
        (width (abs w))
        (top (if (minusp h) (+ y h) y))
        (heigth (abs h)))
     (om-draw-rect left top width heigth)))

:foreground *om-select-color*

(defun om-draw-hilite-rect (x y w h &optional xor) 
  (gp::with-graphics-state (*curstream* :foreground *om-select-color-alpha*)
    (om-fill-rect x y (- w 1) (- h 1))))


(defmacro om-with-pen ((view &key mode pattern (line :black-line) (size 1)) &body body)
  `(let () ,@body))

(defmacro om-with-line-size (size &body body)
  `(gp::with-graphics-state (*curstream* :thickness ,size)
    ,@body))

(defmacro om-with-hilite-foreground (&body body)
  `(gp::with-graphics-state (*curstream* :foreground *om-select-color-alpha*)
    ,@body))

(defmacro om-with-dashline (&body body)
  `(gp::with-graphics-state (*curstream* :dashed t :dash '(2 2))
    ,@body))



(defun om-draw-rect-outline (x y w h &optional (pensize 1))
  (gp::with-graphics-state (*curstream* :thickness pensize)
    (om-draw-rect x y (- w 1) (- h 1))))



  
