;OpenMusic
;
;Copyright (C) 1997, 1998, 1999, 2000 by IRCAM-Centre Georges Pompidou, Paris, France.
; 
;This program is free software; you can redistribute it and/or
;modify it under the terms of the GNU General Public License
;as published by the Free Software Foundation; either version 2
;of the License, or (at your option) any later version.
;
;See file LICENSE for further informations on licensing terms.
;
;This program is distributed in the hope that it will be useful,
;but WITHOUT ANY WARRANTY; without even the implied warranty of
;MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;GNU General Public License for more details.
;
;You should have received a copy of the GNU General Public License
;along with this program; if not, write to the Free Software
;Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
;
;Authors: Gerard Assayag and Augusto Agon


;DocFile
;This file contains the definition of the basic boxes in OM
;(i.e. sequence, omand, omor, omif, etc)
;For each box we subclass the OMboxCall class in order to redefine :
;      the evaluation      (special-value method) or
;      the persistance     (omng-save method) or
;      the code generation (gen-code method), etc.
;Last Modifications :
;18/10/97 first date.
;DocFile

(in-package :om)

;===============================================================================
;CONTROL
;===============================================================================

;--------------SEQUENCE--------------------

(defclass box-seq-call (OMBoxcall)  
   ((numouts :initform 1 :accessor numouts))
   (:documentation "This is the class for the method exception box SEQUENCE.#enddoc#
#seealso# (OMBoxcall) #seealso#
#numouts# The number of outputs in Sequence boxes can change dynamicly, this slots contains this number. #numouts#"))

(defmethod OpenEditorframe ((self box-seq-call)) 
   (not (dialog-message "Compiled Function Sequence")))

(defmethod get-boxcallclass-fun ((self (eql 'sequence))) 'box-seq-call)

(om-with-redefinitions
 (defmethod* sequence  ((patch t) &rest oppatch) :numouts 1 
   :initvals '(nil) :indoc '("patch")
   :icon 161
   :doc "Evaluate sequentially a series of subpatches.
This function may have as many optional inputs as needed. 
Subpatches connected to these inputs are sequentially evaluated.
sequence has as many outputs as inputs. To recover the result of the nth subpatch
after the sequential evaluation, evaluate the nth output (all the subpatches will be evaluated anyway)."
   (let* ((thelist (append (list patch) oppatch))
          (numout (car (last thelist)))
          (thelist (butlast thelist)))
     (nth numout thelist))))

(defmethod call-gen-code ((self box-seq-call) numout)
   `(sequence ,.(decode self) , numout))

(defmethod gen-code-call ((self box-seq-call))
   `(values ,.(decode self)))

(defmethod special-value ((self  box-seq-call) &optional (args nil))
  (values-list args))

(defmethod do-delete-one-input-extra ((self box-seq-call))
   "When you delete an optional input from the sequence boxes you must delete one output also."
   (setf (numouts self) (- (numouts self) 1))
   t)

(defmethod do-add-one-input-extra ((self box-seq-call))
   "When you add an optional input to the sequence boxes you must add one output too."
   (setf (numouts self) (+ (numouts self) 1))
   t)


(defmethod omNG-save ((self box-seq-call) &optional (values? nil))
  (let* ((inputs (mapcar #'(lambda (input) (omNG-save input values?)) (inputs self)))
         (value (when values? (omNG-save (value self) values?))))
    `(om-load-seqbox  ,(name self) ',(reference self) ',inputs ,(om-save-point (frame-position self)) 
                      ,(om-save-point (frame-size self)) ,value ,(allow-lock self) ,(numouts self))))

;Used to restore a sequence box.
(defmethod om-load-seqbox (name reference inputs position size value lock numouts)
   (let ((newbox (omNG-make-new-boxcall (mk-object-refer 'genfun  reference) (om-correct-point position) name)))
     (setf (frame-size newbox) (om-correct-point size))
     (setf (numouts newbox) numouts)
     (setf (inputs newbox) (mapcar #'(lambda (input) (eval input)) inputs))
     (setf (value newbox) value)
     (setf (allow-lock newbox) lock)
     newbox))


;--------------CALL-NEXT-METHOD-------------------------

(defclass box-next-call (OMBoxcall) ()
   (:documentation "This is the class for the method exception box CALLNEXT-METHOD.
OM Callnext-method function is the equivalent of the CLOS fun call-next-method.#enddoc#
#seealso# (OMBoxcall) #seealso#"))

(defmethod OpenEditorframe ((self box-next-call)) 
   (not (dialog-message "Compiled Function callnext-method")))

(defmethod get-boxcallclass-fun ((self (eql 'callnext-method))) 'box-next-call)

(defmethod* callnext-method  () :numouts 1 :initvals nil  
   :doc "call next method.
To use in a method definition. Will call the method with the same name belonging
to the direct upper-classe(s) of the  method argument(s). The arguments
are automatically passed to it so there is no input." 
   :icon 147
   nil)

(defmethod gen-code ((self box-next-call) numout)
  (declare (ignore numout))
  `(call-next-method))

(defmethod special-value ((self  box-next-call) &optional (args nil))
  (declare (ignore args))
  (dialog-message "Do not evaluate this box")
  (om-abort))

;--------------REPEAT-N-------------------------

(defclass box-repeat-n-call (OMBoxcall)  
   ()
   (:documentation "This is the class for the method exception box REPEAT-N.#enddoc#
#seealso# (OMBoxcall) #seealso#"))

(defmethod get-boxcallclass-fun ((self (eql 'repeat-n))) 'box-repeat-n-call)

(defmethod OpenEditorframe ((self box-repeat-n-call)) 
  (not (dialog-message "Compiled Function repeat-n")))

(defmethod* repeat-n  ((self t) (num integer)) :numouts 1 :initvals '(nil 0) :indoc '("patch" "times")
  :doc "repeats n times the evaluation of <self> and collects the n results into a list. " :icon 181
  (let (rep)
    (loop for i from 1 to num do
          (push self rep))
    (reverse rep)))

(defvar *start-repeat-generation* nil)
(defvar *repeat-ev-once-list* nil)

(defmethod call-gen-code ((self box-repeat-n-call) numout)
   (declare (ignore numout))
   (let ((times (gen-code (second (inputs self)) 0))
         (*start-repeat-generation* t) code1 rep)
     (setf *repeat-ev-once-list* nil)
     (setf code1 (gen-code (first (inputs self)) 0))
     (setf rep `(progn
                  ,.*repeat-ev-once-list*
                  (loop for i from 1 to ,times 
                        collect ,code1)))
     (setf *repeat-ev-once-list* nil)
     rep))

(defmethod gen-code-call ((self box-repeat-n-call))
   (call-gen-code self 0))
  
(defmethod special-value ((self  box-repeat-n-call) &optional (args nil))
  (declare (ignore numout))
  (let* ((times (second args))
         (rep (if (zerop times) nil (list (first args)))))
    (loop for i from 1 to (- times 1) do
          (push (omNG-box-value (first (inputs self))) rep))
    (reverse rep)))
 
;--------------OMIF-------------------------

(defclass OMboxif (OMBoxcall) ()
   (:documentation "This is the class for the method exception box OMIF.#enddoc#
#seealso# (OMBoxcall) #seealso#"))

(defmethod get-boxcallclass-fun ((self (eql 'omif))) 'OMBoxif)

(defmethod OpenEditorframe ((self OMBoxif)) 
   (not (dialog-message "Compiled Function OMif")))

(defmethod* omif ((test t) (action t) &optional else) 
   :numouts 1 
   :initvals '(nil nil nil) 
   :indoc '("if" "then" "else")
   :doc "if <test> then <action> else <else>.
If the evaluation of the subpatch connected to <test> is not NIL, 
evaluates the subpatch connected to <action> otherwise evaluates <else>." 
   :icon 180
   (if test action else))


(defmethod call-gen-code ((self OMBoxif) numout)
   (declare (ignore numout))
   `(if ,(gen-code (first (inputs self)) 0) ,(gen-code (second (inputs self)) 0) ,(gen-code (third (inputs self)) 0)))

(defmethod gen-code-call ((self OMBoxif))
   (call-gen-code self 0))
 
(defmethod omNG-box-value ((self OMBoxif) &optional (numout 0))
   (declare (ignore numout))
   (cond
    ((equal (allow-lock self) "l") (special-lambda-value self 'omif))
    ((and (equal (allow-lock self) "x") (value self)) (nth 0 (value self)))
    ((and (equal (allow-lock self) "&") (ev-once-p self)) (nth 0 (value self)))
    (t (let ((test (omNG-box-value (first (inputs self))))
             rep)
         (if test
           (setf rep (list (omNG-box-value (second (inputs self)))))
           (setf rep (list (when (third (inputs self)) (omNG-box-value (third (inputs self)))))))
         (when (equal (allow-lock self) "&")
           (setf (ev-once-p self) t)
           (setf (value self) rep))
         (when (equal (allow-lock self) "x")
           (setf (value self) rep))
         (nth 0 rep)))))

;--------------CONDITIONAL-------------------------

(defclass OMBoxcond (OMBoxcall) ()
   (:documentation "This is the class for the method exception box CONDITIONAL.#enddoc#
#seealso# (OMBoxcall) #seealso#"))

(defmethod get-boxcallclass-fun ((self (eql 'conditional))) 'OMBoxcond)

(defmethod OpenEditorframe ((self OMBoxcond)) 
   (not (dialog-message "Compiled Function conditional")))

(defmethod* conditional ((test t) &rest addtest) 
   :numouts 1 
   :initvals '(nil t) 
   :indoc '("test" "add-test")
   :doc "Eval from left to right while not true.
This functions works as a OR." 
   :icon 180
   (declare (ignore addtest))
   t)

(defmethod call-gen-code ((self OMBoxcond) numout)
   (declare (ignore numout))
   `(or ,.(decode self)))

(defmethod gen-code-call ((self OMBoxcond))
   (call-gen-code self 0))
 
(defmethod omNG-box-value ((self OMBoxcond) &optional (numout 0))
   (declare (ignore numout))
   (cond
    ((equal (allow-lock self) "l") (special-lambda-value self 'conditional))
    ((and (equal (allow-lock self) "x") (value self)) (nth 0 (value self)))
    ((and (equal (allow-lock self) "&") (ev-once-p self)) (nth 0 (value self)))
    (t (let (rep)
         (loop for item in (inputs self)
               while (not rep) do
               (setf rep (omNG-box-value item)))
         (when rep (setf rep (list rep)))
         (when (equal (allow-lock self) "&")
           (setf (ev-once-p self) t)
           (setf (value self) rep))
         (when (equal (allow-lock self) "x")
           (setf (value self) rep))
         (nth 0 rep)))))

;--------------Initialize-Instance---------------------

(defclass OMBoxCallNextInit (OMBoxcall) 
   ((whichclass :initform nil :initarg :whichclass :accessor whichclass))
   (:documentation "This is the class for the method exception box INIT-INSTANCE.#enddoc#
#seealso# (OMBoxcall) #seealso#
#whichclass# Say the class of the instance which is being initialized. #whichclass#"))

(defmethod get-boxcallclass-fun ((self (eql 'init-instance))) 'OMBoxCallNextInit)
(defmethod get-frame-class ((self OMBoxCallNextInit)) 'nonbuttonboxframe)

(defmethod OpenEditorframe ((self OMBoxCallNextInit)) 
   (not (dialog-message "Compiled Function init-instance")))

(defmethod* init-instance ((self t) &rest args) 
   (declare (ignore args))
   :initvals '(nil nil) 
   :indoc '("self" "slot")
   :doc "Used in the initialise-instance method for a class." 
   :icon 147
   nil)

;construct
(defun omNG-make-boxcallnext (class posi name)
   "This function makes a init-instance box called <name> for the class <class> at <posi>."
   (let* ((inputs (get-inputs-from-class (class-name class)))
          rep)
     (setf rep (eval `(make-instance 'OMBoxCallNextInit 
                        :name ,name
                        :reference 'init-instance 
                        :icon 147
                        :whichclass ',(class-name class)
                        :inputs (list ,.inputs))))
     (setf (frame-position  rep) (borne-position posi))
     (setf (frame-size rep) (om-make-point (+ 5 (* (length inputs) 55)) 45))
     (setf (allow-lock  rep) "&")
     rep))


(defmethod gen-code ((self OMBoxCallNextInit) numout)
   (declare (ignore numout)) 'self)

(defmethod redef-boxcallnext ((self OMBoxCallNextInit))
   "Called when you redefine a class and it had an init method."
   (let* ((new-inputs (get-inputs-from-class (whichclass self))))
     (mapc #'(lambda (oldin newin) 
               (setf (connected? newin) (connected? oldin))) 
           (inputs self) new-inputs)
     (setf (inputs self) new-inputs)))

 
(defmethod omNG-box-value ((self OMBoxCallNextInit) &optional (numout 0))
   (declare (ignore numout))
   (not (dialog-message "You can not evaluate the function init-instance")))

(defmethod omNG-save ((self OMBoxCallNextInit) &optional (values? nil))
   (declare (ignore values?))
   `(om-load-boxcallnext ',(whichclass self) ,(om-save-point (frame-position self)) ,(frame-name self)))

(defun om-load-boxcallnext (class posi name)
   (omNG-make-boxcallnext (find-class class nil) posi name))


;--------------OMOR-------------------------

(defclass ORboxCall (OMBoxcall) ()
   (:documentation "This is the class for the method exception box OMOR.#enddoc#
#seealso# (OMBoxcall) #seealso#"))

(defmethod get-boxcallclass-fun ((self (eql 'omor))) 'ORboxCall)

(defmethod OpenEditorframe ((self ORboxCall)) 
   (not (dialog-message "Compiled Function omor")))

(defmethod* OMor  ((self t) &rest rest) :numouts 1 :initvals '(nil nil) :indoc '("patch" "add patch")
   :doc "Logical or.
omor yields the value T (true) if one at least among the subpatches connected
to it evaluates to T. Accepts as many optional inputs as needed.
Generally used as input to  an omIf." 
   :icon 219
   (eval `(or ,self ,.rest)))

(defmethod omNG-box-value ((self ORboxCall) &optional (numout 0))
   (cond
    ((equal (allow-lock self) "l") (special-lambda-value self 'omor))
    ((and (equal (allow-lock self) "x") (value self)) (nth numout (value self)))
    ((and (equal (allow-lock self) "&") (ev-once-p self)) (nth numout (value self)))
    (t (let ((rep (omNG-box-value (first (inputs self)))))
         (loop while (not rep)
               for item in (cdr (inputs self)) do
               (setf rep (omNG-box-value item))) 
         (when (equal (allow-lock self) "&")
           (setf (ev-once-p self) t)
           (setf (value self) (list rep)))
         (when (equal (allow-lock self) "x")
           (setf (value self) (list rep)))
         rep))))

(defmethod call-gen-code ((self ORboxCall) numout)
   (declare (ignore numout))
   `(or ,.(decode self)))

(defmethod gen-code-call ((self ORboxCall))
   (call-gen-code self 0))
  

(defmethod special-value ((self  ORboxCall) &optional (args nil))
   (declare (ignore numout))
   (eval `(or ,.args)))

;--------------OMAND-------------------------

(defclass ANDboxCall (OMBoxcall) ()
   (:documentation "This is the class for the method exception box OMand.
We must use an specil subclass of OMBoxcall because and do not eval all inputs.#enddoc#
#seealso# (omand OMBoxcall  get-boxcallclass-fun) #seealso#"))

(defmethod get-boxcallclass-fun ((self (eql 'omand))) 'ANDboxCall)

(defmethod OpenEditorframe ((self ANDboxCall)) 
  (not (dialog-message "Compiled Function omand")))

(defmethod* omand  ((self t) &rest rest) :numouts 1 :initvals '(nil nil) :indoc '("patch" "add patch")
   :doc "Logical and.
omand yields the value T (true) if all the subpatches connected
to it evaluates to T. Accepts as many optional inputs as needed.
Generally used as input to  an omIf."
   :icon 218
   (eval `(and ,self ,.rest)))

(defmethod omNG-box-value ((self ANDboxCall) &optional (numout 0))
   (cond
    ((equal (allow-lock self) "l") (special-lambda-value self 'omand))
    ((and (equal (allow-lock self) "x") (value self)) (nth numout (value self)))
    ((and (equal (allow-lock self) "&") (ev-once-p self)) (nth numout (value self)))
    (t (let ((rep (omNG-box-value (first (inputs self)))))
         (loop while rep
               for item in (cdr (inputs self)) do
               (setf rep (omNG-box-value item))) 
         (when (equal (allow-lock self) "&")
           (setf (ev-once-p self) t)
           (setf (value self) (list rep)))
         (when (equal (allow-lock self) "x")
           (setf (value self) (list rep)))
         rep))))


(defmethod call-gen-code ((self ANDboxCall) numout)
   (declare (ignore numout))
   `(and ,.(decode self)))

(defmethod gen-code-call ((self ANDboxCall))
   (call-gen-code self 0))
  

(setf *predicate-package* (omNG-protect-object (omNG-make-new-package "Predicates")))
(setf *control-package* (omNG-protect-object (omNG-make-new-package "Control")))
 
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'omloop)))
(addGenFun2Pack 'omloop *control-package*) 
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'repeat-n)))
(addGenFun2Pack 'repeat-n *control-package*)
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'callnext-method)))
(addGenFun2Pack 'callnext-method *control-package*)
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'sequence)))
(addGenFun2Pack 'sequence *control-package*)
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'omif)))
(addGenFun2Pack 'omif *control-package*)
;;;(omNG-add-element *control-package* (omNG-protect-object (fdefinition 'conditional)))
(addGenFun2Pack 'conditional *control-package*)

;;;(omNG-add-element *predicate-package* (omNG-protect-object (fdefinition 'omand)))
(addGenFun2Pack 'omand *predicate-package*)
;;;(omNG-add-element *predicate-package* (omNG-protect-object (fdefinition 'omor)))
(addGenFun2Pack 'omor *predicate-package*)


;===============================================================================
;DATA
;===============================================================================

(defun is-om-slot? (class slot-symb)
   (let ((slotlist (get-all-slots-of-class class)))
     (member (string slot-symb) slotlist :test 'string-equal :key 'name)))
     
(defmethod* get-slot ((object t) (slot symbol))
   :initvals '(nil nil) 
   :indoc '("object" "slot") 
   :icon 237
   :doc
   "An object slot inspection. The first input must be an object-instance 
(such as chord objects or breakpoint functions). 
The second input is a slot-name. Returns the corresponding value of the 
chosen slot. 
Do not use on OM predefined object, use rather the get-set-slot mechanism by
shift-dragging a class icon from a package to a patch window."
   (if (omclass-p (class-of (class-of object)))
     (if (is-om-slot? (type-of object) slot)
       (slot-value object (internp (string slot) (symbol-package (type-of object))))
       (om-beep-msg (string+ (string slot) " is not an OM slot of the class " (string (type-of object)) ".")))
     (om-beep-msg (string+ (string (type-of object)) " is not an OM class."))))

(defmethod* get-slot ((object list) (slot symbol))
   (loop for item in object 
         collect  (get-slot item slot)))

(defmethod* set-slot ((object t) (slot symbol) (value t))
   :initvals '(nil nil nil) 
   :indoc '("object" "slot" "value") 
   :icon 335
   :doc
   "An object slot modification. The first input must be an object-instance.
The second input is  a slot-name . Third input is the 
corresponding new value. Returns the modified object. 
Warning: This operation is potentially dangerous. You should know what you 
are doing when changing object slot's values.
Do not use on OM predefined object, use rather the get-set-slot mechanism by
shift-dragging a class icon from a package to a patch window."
   (if (omclass-p (class-of (class-of object)))
     (if (is-om-slot? (type-of object) slot)
        (setf (slot-value  object (internp (string slot) (symbol-package (type-of object)))) value)
       (om-beep-msg (string+ (string slot) " is not an OM slot of the class " (string (type-of object)) ".")))
     (om-beep-msg (string+ (string (type-of object)) " is not an OM class."))))
  
 
;===============================================================
;Instances of this class can be used as global variable

(defclass* store ()
   ((value :initform nil :initarg :value :accessor value))
   (:icon 215)
   (:documentation "Use this class to simulate variables"))

  
;===============================================================
;copy

(defmethod omNG-copy ((self t))
   "Return an expresion that produce a copy of 'self' when it is valued."
   (cond
    ((om-color-p self) (om-copy-color self))
    ((om-point-p self) (om-copy-point self))
    ((omclass-p  (class-of self))
     (let* ((theclass (class-name (class-of self)))
            (slots (mapcan #'(lambda (slot)
                               (list (string2initarg (string (first slot))) 
                                     (omNG-copy (slot-value self (first slot))))) 
                           (get-initargs-of-class theclass))))
       `(make-instance ',theclass ,.slots)))
    ((equal (class-name (class-of (class-of self))) 'standard-class)
     (let* ((theclass (class-name (class-of self)))
            (slots (mapcan #'(lambda (slot)
                               (list `(setf (slot-value copy ',(first slot)) 
                                            ,(omNG-copy (slot-value self (first slot)))))) 
                           (get-init-slots-of-class theclass))))
       `(let ((copy (make-instance ',theclass)))
          ,.slots
          copy)))
    (t self)))


(defmethod omNG-copy ((self function)) self)

(defmethod omNG-copy ((self hash-table)) (omng-save self))

(defmethod omNG-copy ((self null))  nil)

(defmethod omNG-copy ((self symbol)) `',self)

(defmethod omNG-copy ((self list)) `(list ,@(mapcar #'omNG-copy self)))

(defmethod omNG-copy ((self cons)) 
  (if (quoted-form-p self)
    `',(omng-copy (second self))
    (if (null (cdr (last self)))
       (call-next-method)
       `(cons ,(omng-save (first (last self))) ,(omng-save (cdr (last self)))))))



(defmethod* clone ((self t)) 
   :icon 205 
   :indoc '("The object to copy")
   :doc "Make a copy of an object"
   (eval (omng-copy self)))

(defmethod* clone ((self simple-container))
  (copy-container self))
 
(setf *data-package* (omNG-protect-object (omNG-make-new-package "Data")))
;;;(omNG-add-element *data-package* (omNG-protect-object (fdefinition 'set-slot)))
(addGenFun2Pack 'set-slot *data-package*)
;;;(omNG-add-element *data-package* (omNG-protect-object (fdefinition 'get-slot)))
(addGenFun2Pack 'get-slot *data-package*)
;;;(omNG-add-element *data-package* (omNG-protect-object (fdefinition 'clone)))
(addGenFun2Pack 'clone *data-package*)
;;;(omNG-add-element *data-package* (omNG-protect-object (find-class 'store)))
(addClass2Pack 'store *data-package*)


;--------------------------FOR maquettes

(defmethod* TemporalBoxes ((self OMMaquette))
   :icon 326
   :indoc '("A maquette")
   "Get de temporal boxes in <self>." 
   (loop for item in (boxes self)
         when (boxtempobj-p item)
         collect item))

(defmethod* addBox2Maquette ((self temporalbox) (maquette ommaquette))
   :icon 328
   :initvals '(nil nil)
   :indoc '("The temporal object" "the maquette")
   :doc "Put <self> ( a temporal box or a list of temporal boxes ) in the maquette <maquette>"
   (if (not (editorframe maquette))
     (omng-add-element maquette self)
     (omg-add-element (editorframe maquette) (make-frame-from-callobj self)))
   self)

(defmethod* addBox2Maquette ((self list) (maquette ommaquette))
   (loop for item in self 
         collect (addBox2Maquette item maquette)))

(defmethod* removeTemporalBox ((self temporalbox))
   :icon 327
   :initvals '(nil)
   :indoc '("the temporal object")
   :doc "Remove <self> ( a temporal box or a list of temporal boxes ) from its container if it has one."
   (let ((maquette (mycontainer self)))
     (when maquette
       (if (not (editorframe maquette))
         (omng-remove-element maquette self)
         (omg-remove-element (editorframe maquette) (car (frames self)))))
     self))

(defmethod* removeTemporalBox ((self list))
   (loop for item in self 
         collect (removeTemporalBox item)))

(defmethod* get-maquette ((self temporalbox))
   :icon 333
   :indoc '("A temporal box")
   "Get de maquette containing the temporal box <self>." 
   (mycontainer self))

;;;;=====================================
(defmethod! set-eval-func ((self OMMaquette) (func OMMaquette)) nil)
(defmethod! set-eval-func ((self OMMaquette) (func t)) nil)

(defmethod! set-eval-func ((self OMMaquette) (func null)) 
  (setf (eval-func self) nil))

(defmethod! set-eval-func ((self OMMaquette) (func OMPatch))
  (setf (eval-func self) (omng-make-new-boxcall func (om-make-point 0 0) "maqfunc"))
  (unless (loaded? func) (load-patch func))
  (when (editorframe self)
    (om-invalidate-view (editor (editorframe self)))) 
  self)

(defmethod! set-eval-func ((self OMMaquette) (func OMPatchAbs))
  (setf (eval-func self) (omng-make-new-boxcall (eval (omng-copy func)) (om-make-point 0 0) "absfunc"))
  (unless (loaded? func) (load-patch func))
  (when (editorframe self)
    (om-invalidate-view (editor (editorframe self)))) 
  self)

(defmethod! set-eval-func ((self OMMaquette) (func OMGenericFunction))
  (setf (eval-func self) (omng-make-new-boxcall func (om-make-point 0 0) "maqfunc"))
  (when (editorframe self)
    (om-invalidate-view (editor (editorframe self)))) 
  self)

(defmethod! set-eval-func ((self OMMaquette) (func OMLispFun))
  (setf (eval-func self) (omng-make-new-boxcall func (om-make-point 0 0) "maqfunc"))
  (when (editorframe self)
    (om-invalidate-view (editor (editorframe self)))) 
  self)

(defmethod! set-eval-func ((self OMMaquette) (func symbol))
  (when (fboundp func)
    (set-eval-func self (omNG-make-new-lispfun func))))


;;;========================================


(defmethod! get-tempobjs-value ((self list) &optional (test nil))
  (let ((rep nil))
    (loop for tempbox in self do
          (when (boxtempobj-p tempbox)
            (if (or (not test) (funcall test (get-mus-ob tempbox)))
              (push (get-mus-ob tempbox) rep))))
    (reverse rep)))

(defmethod! get-tempobjs ((self list) &optional (test nil))
  (let ((rep nil))
    (loop for tempbox in self do
          (when (boxtempobj-p tempbox)
            (if (or (not test) (funcall test tempbox))
              (push tempbox rep))))
    (reverse rep)))


(defmethod! test-value-class ((self Temporalbox) class)
  (equal class (class-name (class-of (get-mus-ob self)))))

(defmethod! test-obj-class ((self t) class)
  (equal class (class-name (class-of self))))


;--------------------------------------------------------------------

(defmethod make-one-instance ((self Temporalbox) &rest slots-vals)
   (setf slots-vals (apply 'pretraitement (cons self slots-vals)))
   (apply 'make-a-tempobj slots-vals))

#|
(defmethod make-a-tempobj ((offset integer) (extend integer) (color t) 
                             value (posy integer) (strech number) (sizey integer) store refer)
   (unless refer (setf refer (make-instance 'OMTemporalPatch :name "tempobj")))
   (let* ((newobj (omNG-make-tempobj refer (om-make-point 0 0) "tempobj")))
     (setf (offset newobj) offset)
     (setf (extend newobj) extend)
     (setf (sizey newobj) sizey)
     (setf (posy newobj) posy)
     (setf (free-store newobj) store)
     (when value (setf (value newobj) value))
     (setf (strech-fact newobj) strech)
     (setf (colorframe newobj) color)
     newobj))
|#

(defmethod make-a-tempobj ((offset integer) (extend integer) (color t) 
                             value (posy integer) (strech number) (sizey integer) store refer &optional mute lock)
   (unless refer (setf refer (make-instance 'OMPatchAbs :name "tempobj")))
   (let* ((newobj (omNG-make-tempobj refer (om-make-point 0 0) "tempobj")))
     (setf (offset newobj) offset)
     (setf (extend newobj) extend)
     (setf (sizey newobj) sizey)
     (setf (posy newobj) posy)
     (setf (free-store newobj) store)
     (when value (setf (value newobj) value))
     (setf (strech-fact newobj) strech)
     (setf (colorframe newobj) color)
     (setf (mute newobj) mute)
     (setf (lock newobj) lock)
     newobj))

;these functions are kept by compatibility with old patches but they are not putted in the package

(defmethod* starttime ((self temporalbox) &optional (val nil))
  :icon 239
  :numouts 2
  :initvals '(nil nil)
  :indoc '("the temporal object" "if you use this arg you make a set")
  :doc "get or set the start time of the temporal object"
  (if (null val) (values self (offset self))
      (setf (offset self) val))
  (values self (offset self)))

(defmethod* objdur ((self temporalbox) &optional (val nil))
  :icon 239
  :numouts 2
  :initvals '(nil nil)
  :indoc '("the temporal object" "if you use this arg you make a set")
  :doc "get or set the duration of the temporal object"
  (cond
   ((null val) (values self (extend self)))
   ((and (integerp val) (> val 0))
    (setf (extend self) val)
    (values self (extend self)))
   (t
    (om-beep-msg "Bad value in the function objdur"))))

(defmethod* mymaquette ((self temporalbox))
  :icon 239
  :initvals '(nil)
  :indoc '("the temporal object" )
  :doc "get the maquette which contains self"
  (mycontainer self))



(defmethod* put-in-maq ((self temporalbox) (maquette ommaquette))
  :icon 239
  :initvals '(nil nil)
  :indoc '("the temporal object" "the maquette")
  :doc "put in the maquette"
  (if (not (editorframe maquette))
    (omng-add-element maquette self)
    (omg-add-element (editorframe maquette) (make-frame-from-callobj self)))
  self)

 
(defmethod* objfree-store ((self temporalbox) &optional control)
  :icon 239
  :numouts 2
  :initvals '(nil nil)
  :indoc '("the temporal object" )
  :doc "get or set the freq temporal object"
  (cond
   ((null control) (values self (free-store self)))
   (t (setf (free-store self) control)
      (values self (free-store self)))))

(defmethod* eval-tempobj ((self temporalbox))
  :icon 239
  :initvals '(nil)
  :indoc '("the temporal object" )
  :doc "get or set the freq temporal object"
  (cond
   ((car (frames self)) (eval+redraw (car (frames self))))
   (t (omng-box-value self))))



(setf *maquette-package* (omNG-protect-object (omNG-make-new-package "Maquette")))
;;;(omNG-add-element *maquette-package* (omNG-protect-object (find-class 'TemporalBox)))
(AddClass2Pack 'TemporalBox *maquette-package*)

(AddGenFun2Pack '(addBox2Maquette removeTemporalBox TemporalBoxes get-maquette) *maquette-package*)


(defmethod! om-shell ((str string))
  :icon 642
  (om-cmd-line str))

(AddGenFun2Pack '(om-shell) *control-package*)