;*****************************************************************************
;-------| DG SYSTEM
;-------| This file is: $LLdg/spectrum.ll
;-------| Version V1.0: Jan 22, 1990
;-------| By Marco Stroppa
;-------| Copyright 1990 IRCAM
;*****************************************************************************

; THIS FILES CONTAINS VARYING FUNCTIONS TO GENERATE HARMONIC, SHIFTED,
;	AND STRETCHED SPECTRA ACCORDING TO MC ADAMS' FORMULAS

(in-package chroma)
;-----------------------------------------------------------------------------
; AVAILABLE FUNCTIONS
;	sp
;	spsh
;	spst
;	spsht
;	xpose / xp-int
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (sp F0 NP [STON])
;	generate a harmonic spectrum
;	F0 = fundamental frequency
;	NP = number of components to generate
;	   it may be:
;		a NUMBER : all the components from 1 to NP will be generated
;		a LIST containing a list of the scalers of f0
;	STON = % of detuning applied when computing each component [0-1]
; EX: (sp 440 10), (sp 440 10 0.06), (sp 440 '(2 3.3 5.1 7) 0.01)
(defun sp (f0 np &rest ston)
  (let ((ston (ifn ston 0 (car ston)))
        (result ()) )
    (if (numberp np)
      (let ((cnt 0))
        (loop while (<= (incf cnt) np)
          do (newl result (* (+ 1 (ran ston))
                          cnt
                          f0)) ))
      (loop while np
        do (let ((curr (nextl np)) )
          (newl result (* (+ 1 (ran ston))
                          curr
                          f0)) )) )
    (nreverse result)) )
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (spsh F0 NP SH [STON])
;	generate a shifted spectrum
;	F0 = fundamental frequency
;	NP = number of components to generate
;	   it may be:
;		a NUMBER : all the components from 1 to NP will be generated
;		a LIST containing a list of the scalers of f0
;	SH = shifting factor (% of f0)
;	STON = % of detuning applied when computing each component [0-1]
; EX: (spsh 440 10 0.06), (spsh 440 10 0.1 0.06)
;	(spsh 440 '(2 3.3 5.1 7) 0.5 0.01)
(defun spsh (f0 np sh &rest ston)
  (let ((ston (ifn ston 0 (car ston)))
        (result ())
        (shift (* f0 sh)) )
    (if (numberp np)
      (let ((cnt 0))
        (loop while (<= (incf cnt) np)
          do (newl result (+  (* (+ 1 (ran ston))
                              cnt
                              f0)
                           shift))) )
      (loop while np
        do (let ((curr (nextl np)) )
          (newl result (+  (* (+ 1 (ran ston))
                              curr
                              f0)
                           shift))) ))
    (nreverse result)) )
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (spst F0 NP ST [STON])
;	generate a stretched/compressed spectrum
;	F0 = fundamental frequency
;	NP = number of components to generate
;	   it may be:
;		a NUMBER : all the components from 1 to NP will be generated
;		a LIST containing a list of the scalers of f0
;	ST = stretching/compressing factor (2.0 = octave)
;	STON = % of detuning applied when computing each component [0-1]
; EX: (spst 440 10 2.1), (spst 440 10 1.76 0.06)
;	(spst 440 '(2 3.3 5.1 7) 2.5 0.01)
(defun spst (f0 np st &rest ston)
  (let ((ston (ifn ston 0 (car ston)))
        (result ())
        (stretch (/ (log st) (log 2.0))) )
    (if (numberp np)
      (let ((cnt 0))
        (loop while (<= (incf cnt) np)
          do (newl result (* (+ 1 (ran ston))
                          (expt cnt stretch)
                          f0)) ))
      (loop while np
        do (let ((curr (nextl np)) )
          (newl result (* (+ 1 (ran ston))
                          (expt curr stretch)
                          f0)) )) )
    (nreverse result)) )
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (spsht F0 NP SH ST [STON])
;	generate a shifted spectrum
;	F0 = fundamental frequency
;	NP = number of components to generate
;	   it may be:
;		a NUMBER : all the components from 1 to NP will be generated
;		a LIST containing a list of the scalers of f0
;	SH = shifting factor (% of f0)
;	ST = stretching/compressing factor (2.0 = octave)
;	STON = % of detuning applied when computing each component [0-1]
; EX: (spsht 440 10 2.1), (spsht 440 10 1.76 0.06)
;	(spsht 440 '(2 3.3 5.1 7) 2.5 0.01)
(defun spsht (f0 np sh st &rest ston)
  (let ((ston (ifn ston 0 (car ston)))
        (result ())
        (shift (* f0 sh))
        (stretch (/ (log st) (log 2.0))) )
    (if (numberp np)
      (let ((cnt 0))
        (loop while (<= (incf cnt) np)
          do (newl result (+  (* (+ 1 (ran ston))
                              (expt cnt stretch)
                              f0)
                           shift)) ))
      (loop while np
        do (let ((curr (nextl np)) )
          (newl result (+  (* (+ 1 (ran ston))
                              (expt curr stretch)
                              f0)
                           shift)) )) )
    (nreverse result)) )
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (xpose INT SP)
;	transpose a spectrum (list of freq) by an interval INT (scaler of f0)

; EX: (xpose (spsht 440 10 1.76 0.06) -0.5) --> octave below
;     (xpose (spsht 440 10 1.76 0.06) 0.5) --> fifth above
;-----------------------------------------------------------------------------
(defun xpose (sp int)
    (let ((result ()) )
	(loop while sp
	    do (newl result (* int (nextl sp))) )
	(nreverse result)) )
;-----------------------------------------------------------------------------

;-----------------------------------------------------------------------------
; (xp-int INT) / (xp-int-1 INT)
;	generate a transposing interval of INT (in semitones, either positive
;	   or negative) to use with xpose and other functions
;	xp-int will work functions needing with absolute values (e.g. xpose),
;	   whereas xp-int-1 is suited to functions in which the transposing
;	   factor will then be multiplied by the reference values (e.g. the
;	   sh factor of spsh[t])
;	no transposition : xp-int = 1, xp-int-1 = 0
;	one octave : xp-int = 2, xp-int-1 = 1

; EX:     (xpose (spsht 440 10 1.76 0.06) (xp-int 7)) --> fifth above
;-----------------------------------------------------------------------------
(defun xp-int (int)
    (let ((semitone 1.059463095))
	(ifn (listp int)
	    (expt semitone int)
	    (mapcar 'xp-int int)))
  )

(defun xp-int-1 (int)
    (let ((semitone 1.059463095))
	(ifn (listp int)
	    (1- (expt semitone int))
	    (mapcar 'xp-int-1 int)))
  )
;-----------------------------------------------------------------------------
