;*****************************************************************************
;-------| PULS SYSTEM
;-------| This file is: $LLsys/tbl.ll
;-------| Version V1.1: Aug 13, 1991
;-------| By Marco Stroppa
;-------| Copyright 1986 MIT
;*****************************************************************************

(in-package cr)

; PACKAGE TO DEAL WITH ONE- OR TWO-DIMENSIONAL TABLES
;                 ASSOCIATED TYPE NAME: TBL


; DESCRIPTION OF THE DATA STRUCTURE:
;    The structure consists of a one- or two-dimensional table, the elements
;       of which are reached through one or two keys.  Their contents can
;	be anything at all and is retrieved in the same conditions it has been
;	put in.
;    Note: insertion of new els is done at the beginning of the structure
;          there is no consideration for any kind of orders (see lstep.ll).

; CURRENT IMPLEMENTATION:
;    The structure of type TBL is an A-list of sub-tables with type prefix.
;       Each sub-table is an A-list of keys/values and is preceded by its key.


; AVAILABLE FUNCTIONS:

;	CONSTRUCTORS: make_tbl
;	SELECTORS:    lookup_tbl
;		      key-list_tbl
;	MODIFIERS:    insert_tbl
;		      rm_tbl
;	PREDICATES:   is_tbl
;		      is-key_tbl
;		      is-empty_tbl
;	INFO:	      print_tbl
;		      short-print_tbl



; DESCRIPTION OF THE PACKAGE:

;	NAME:		make_tbl  (CONSTRUCTOR)
;	TYPE:		Expr with no arguments
;	CALL:		(make_tbl)
;	FUNCTION:	define an empty structure of type TBL
;	VALUE:		the new structure
;	SOURCE:		$LLsys/tbl.ll

(defun make_tbl ()
    (attach-type 'TBL ()))


;	NAME:		lookup_tbl  (SELECTOR)
;	TYPE:		Expr with 2 or 3 arguments
;	CALL:		(lookup_tbl tbl key1 [key2])
;	FUNCTION:	return the value associated to key1 (and key2 if the
;			   structure is two-dimensional)
;	VALUE:		the above value
;	NOTE:		if there is no key1 (and key2) an error is produced
;	SOURCE:		$LLsys/tbl.ll

(defun lookup_tbl (tbl key1 &rest key2)
  (unless (is_tbl tbl)
    (error-type 'lookup_tbl tbl))
  (when (is-empty tbl)
    (error "Can't find something in an empty table, sir, come on! ~a" 'PRRRR))
  (let ((key2 (ifn key2 key1 (car key2)))
        (subtable (assq key1 (contents tbl))))
    (ifn subtable
         (error "We're sorry, the key ~A you're interested in doesn't exist" key1)
      (let ((record (assq key2 (cdr subtable))))
        (ifn record
             (error "We're sorry, the key ~A you're interested in doesn't exist" key2)
          (cdr record))))))


;	NAME:		key-list_tbl  (PREDICATE)
;	TYPE:		Expr with 1 argument
;	CALL:		(key-list_tbl tbl)
;	FUNCTION:	return a list with all the keys belonging to table tbl
;	VALUE:		the list above (ex. (k1 k2 (k3 k31 k32) k4) )
;	SOURCE:		$LLsys/tbl.ll

(defun key-list_tbl (tbl)
    (unless (is_tbl tbl)
	    (error-type 'key-list_tbl tbl))
    (kltbl (contents tbl)) )

(defun kltbl (ctbl)
   (ifn (car ctbl)
	()
	(cons
	  (let ((key1 (caar ctbl))
	        (rest (cdar ctbl)) )
	    (if (eq key1 (caar rest))
	        key1
	        (append
		  (list key1)
		  (let ((result ()) )
		     (loop while rest
		        do (newl result (car (nextl rest))) )
		     (nreverse result)))
	    ))
	(kltbl (cdr ctbl)))
     ))


;	NAME:		insert_/rm_tbl  (MODIFIERS)
;	TYPE:		Expr with 3 or 4/2 or 3 arguments
;	CALL:		(insert_tbl tbl key1 [key2] val)
;			(rm_tbl tbl key1 [key2])
;	FUNCTION:	respectively insert a new element or delete an already
;			   existing one from a structure of type TBL
;			if the element to insert already exists, only its
;			   value will be changed
;	VALUE:		the symbol 'ok if element exists, () otherwise
;	SOURCE:		$LLsys/tbl.ll

(defun insert_tbl (tbl key1 &rest rest)
  (unless (is_tbl tbl)
    (error-type 'insert_tbl tbl))
  (unless (<= (length rest) 2)
    (error "No insertion, sir!  Can't have so many elements ~a" (cons key1 rest)))
  (let ((key2 (ifn (cdr rest) key1 (nextl rest)))
        (val (car rest))
        (subtable (assq key1 (contents tbl))))
    (ifn subtable
         (progn
           (rplacd tbl
                   (cons (list key1
                               (cons key2 val))
                         (contents tbl)))
           ())
      (let ((record (assq key2 (cdr subtable))))
        (ifn record	
             (progn (rplacd subtable
                            (cons (cons key2 val)
                                  (cdr subtable)))
                    ())
           (rplacd record val)
          )) )
    )'ok)

;(defun rm_tbl (tbl key1 &rest key2)
;  (unless (is_tbl tbl)
;    (error-type 'rm_tbl tbl))
;  (let ((key2 (ifn key2 key1 (car key2)))
;        (subtable (assq key1 (contents tbl))))
;    (ifn subtable
;         (error 'rm_tbl
;                "No deletion, sir!  The key you asked for does not exist"
;                key1)
;     (letn self ((sbtbl subtable))
;            (cond
;             ((null sbtbl)
;              (error 'rm_tbl
;                     "No deletion, sir! The keys you asked for do not exist"
;                     (cons key1 key2)))
;             ((eq (caadr sbtbl) key2)
;              (rplacd sbtbl (cddr sbtbl)))   ; REAL DELETION
;             (t
;              (self (cdr sbtbl))))))
;    (unless (cdr subtable) 		 	; IF SUBTABLE IS NOW EMPTY,
;      (letn self ((tbl tbl))		;    DELETE IT FROM TABLE
;            (if (eq (caadr tbl) key1)
;              (rplacd tbl (cddr tbl))
;              (self (cdr tbl)))))
;    'ok))



(defun rm_tbl (tbl key1 &rest key2)
  (unless (is_tbl tbl)
    (error-type 'rm_tbl tbl))
  (let* ((key2 (ifn key2 key1 (car key2)))
         (subtable (assq key1 (contents tbl))))
    (if (null subtable)
      (error "No deletion, sir!  The key ~a you asked for does not exist" key1)
      (rm_tbl-sub1 subtable key1 key2)) 
    (unless (cdr subtable) 		 	; IF SUBTABLE IS NOW EMPTY,
      (rm_tbl-sub2 tbl key1))
    )'ok)


(defun rm_tbl-sub1 (sbtbl key1 key2)
  (cond
   ((null sbtbl)
    (error "No deletion, sir! The keys ~a you asked for do not exist"(cons key1 key2)))
   ((eq (caadr sbtbl) key2)
    (rplacd sbtbl (cddr sbtbl)))   ; REAL DELETION;
   (t
    (rm_tbl-sub1(cdr sbtbl) key1 key2))))

(defun rm_tbl-sub2 (tbl key1)		;    DELETE IT FROM TABLE
  (if (eq (caadr tbl) key1)
    (rplacd tbl (cddr tbl))
    (rm_tbl-sub2 (cdr tbl) key1)))



;	NAME:		is_tbl  (PREDICATE)
;	TYPE:		Expr with 1 argument
;	CALL:		(is_tbl tbl)
;	FUNCTION:	test whether the argument is a structure of type TBL
;	VALUE:		t or nil according to the test
;	SOURCE:		$LLsys/tbl.ll

(defun is_tbl (tbl)
  (when (is-tagged tbl)
   (eq (pls-type tbl) 'TBL)))


;	NAME:		is-key_tbl  (PREDICATE)
;	TYPE:		Expr with 2/3 arguments
;	CALL:		(is-key_tbl tbl key1 [key2])
;	FUNCTION:	test whether the argument is a valid key for a
;			   structure of type TBL
;	VALUE:		key1 [or key2] or nil according to the test
;	SOURCE:		$LLsys/tbl.ll

(defun is-key_tbl (tbl key1 &rest key2)
    (unless (is_tbl tbl)
	    (error-type 'rm_tbl tbl))
    (let ((key2 (ifn key2 key1 (car key2)))
	  (subtable (assq key1 (contents tbl))))
       (when (and subtable (assq key2 (cdr subtable)) )
	     key2)))


;	NAME:		is-empty_tbl  (PREDICATE)
;	TYPE:		Expr with 1 argument
;	CALL:		(is-empty_tbl tbl)
;	FUNCTION:	test whether a structure of type TBL is empty
;	VALUE:		t or nil according to the test
;	SOURCE:		$LLsys/tbl.ll

(defun is-empty_tbl (tbl)
    (null (contents tbl)))


;	NAME:		print_/short-print_tbl  (INFO)
;	TYPE:		Expr with 1 argument
;	CALL:		(print_tbl tbl)
;			(short-print_tbl tbl)
;	FUNCTION:	nicely or shortely print a structure of type TBL
;	VALUE:		the string 'done
;	SOURCE:		$LLsys/tbl.ll

(defun print_tbl (tbl)
   (terpri)
   (format t "Structure of type : ~a~%" (pls-type tbl))
   (mapc 'prntsbtbl (contents tbl))
   'done)

(defun prntsbtbl (sbtbl)
  (let ((key1 (car sbtbl)))
    (mapc
     (lambda (ssbtbl)
       (if (eq key1 (car ssbtbl))
         (progn
           (prin "  Key1: " key1)
           (format t "    Value: ~a~%" (cdr ssbtbl)))
         (progn (prin "  Key1: " key1 " / Key2: " (car ssbtbl))
               (format t "    Value: ~a~%" (cdr ssbtbl)))))
     (cdr sbtbl))))


(defun short-print_tbl (tbl)
  (format t "~a~%" (pls-type tbl))
  (mapc 'shrtprntsbtbl (contents tbl))
  'done)

(defun shrtprntsbtbl (sbtbl)
  (let ((key1 (car sbtbl)))
    (mapc
     (lambda (ssbtbl)
       (if (eq key1 (car ssbtbl))
         (progn
           (prin "  " key1)
           (format t "    ~a~%" (cdr ssbtbl)))
         (progn(prin "  " key1 " / " (car ssbtbl))
               (format t "    ~a~%" (cdr ssbtbl)))))
     (cdr sbtbl))))
